# StdLib
import os
from random import randint
from time import sleep

from boto3.dynamodb import types
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError
from ddbcache import DDBCache
from moto import mock_dynamodb2
import pytest


@pytest.fixture()
def aws_credentials():  # noqa: PT004
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


@pytest.fixture()
def ddb_cache(aws_credentials):
    with mock_dynamodb2():
        cache = DDBCache(
            table_name="Testing",
            access_key=os.environ["AWS_ACCESS_KEY_ID"],
            secret_key=os.environ["AWS_SECRET_ACCESS_KEY"],
            region="us-east-1",
        )
        key_schema = [
            {
                "AttributeName": "id",
                "KeyType": "HASH",
            },
        ]
        attr_schema = [
            {
                "AttributeName": "id",
                "AttributeType": types.STRING,
            },
        ]
        create_kwargs = {
            "ProvisionedThroughput": {
                "ReadCapacityUnits": 1,
                "WriteCapacityUnits": 1,
            },
        }
        cache.create_table(key_schema, attr_schema, **create_kwargs)
        yield cache


@pytest.fixture()
def init_get(ddb_cache: DDBCache):  # noqa: PT004
    ddb_cache.put_item({"id": "test_get", "data": "test_get"})


@pytest.fixture()
def init_delete(ddb_cache: DDBCache):  # noqa: PT004
    ddb_cache.put_item({"id": "test_delete", "data": "test_delete"})


@pytest.fixture()
def init_update(ddb_cache: DDBCache):  # noqa: PT004
    ddb_cache.put_item({"id": "test_update", "data": "test_update"})


@pytest.fixture()
def init_fetch_cache(ddb_cache: DDBCache):  # noqa: PT004
    ddb_cache.put_cache({"id": "test_fetch_cache", "data": "test_fetch_cache"})


def test_client_error(ddb_cache: DDBCache):
    data = randint(1, 100)  # noqa: S311
    with pytest.raises(ClientError) as err:
        ddb_cache.put_item({"not-id": data, "data": data})
    assert err.value.response["Error"]["Code"] == "ValidationException"
    assert "Missing the key id" in err.value.response["Error"]["Message"]


def test_put(ddb_cache: DDBCache):
    data = randint(1, 100)  # noqa: S311
    ddb_cache.put_item({"id": "test_put", "data": data})
    item = ddb_cache.get_item({"id": "test_put"})
    assert item
    assert item["id"] == "test_put"
    assert item["data"] == data


def test_put_cache(ddb_cache: DDBCache):
    data = randint(1, 100)  # noqa: S311
    ddb_cache.put_cache({"id": "test_put_cache", "data": data})
    item = ddb_cache.get_item({"id": "test_put_cache"})
    assert item
    assert item["id"] == "test_put_cache"
    assert item["data"] == data
    assert item["ttl"]


def test_put_cache_no_ttl(ddb_cache: DDBCache):
    data = randint(1, 100)  # noqa: S311
    ddb_cache.put_cache({"id": "test_put_cache", "data": data}, with_ttl=False)
    item = ddb_cache.get_item({"id": "test_put_cache"})
    assert item
    assert item["id"] == "test_put_cache"
    assert item["data"] == data
    with pytest.raises(KeyError):
        item["ttl"]


def test_fetch_cache(ddb_cache: DDBCache, init_fetch_cache):
    item = ddb_cache.get_item({"id": "test_fetch_cache"})
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    ttl = item["ttl"]
    assert ttl
    sleep(1)
    item = ddb_cache.fetch_cache({"id": "test_fetch_cache"})
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    with pytest.raises(KeyError):
        item["ttl"]
    item = ddb_cache.get_item({"id": "test_fetch_cache"})
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    ttl2 = item["ttl"]
    assert ttl2
    assert ttl != ttl2


def test_fetch_cache_no_ttl_update(ddb_cache: DDBCache, init_fetch_cache):
    item = ddb_cache.get_item({"id": "test_fetch_cache"})
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    ttl = item["ttl"]
    assert ttl
    item = ddb_cache.fetch_cache({"id": "test_fetch_cache"}, with_ttl=False)
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    with pytest.raises(KeyError):
        item["ttl"]
    item = ddb_cache.get_item({"id": "test_fetch_cache"})
    assert item
    assert item["id"] == "test_fetch_cache"
    assert item["data"] == "test_fetch_cache"
    ttl2 = item["ttl"]
    assert ttl2
    assert ttl == ttl2


def test_get(ddb_cache: DDBCache, init_get):
    item = ddb_cache.get_item({"id": "test_get"})
    assert item
    assert item["id"] == "test_get"
    assert item["data"] == "test_get"


def test_delete(ddb_cache: DDBCache, init_delete):
    item = ddb_cache.get_item({"id": "test_delete"})
    assert item
    assert item["id"] == "test_delete"
    assert item["data"] == "test_delete"
    ddb_cache.delete_item({"id": "test_delete"})
    item = ddb_cache.get_item({"id": "test_delete"})
    assert not item


def test_update(ddb_cache: DDBCache, init_update):
    data = randint(1, 100)  # noqa: S311
    item = ddb_cache.get_item({"id": "test_update"})
    assert item
    assert item["id"] == "test_update"
    assert item["data"] == "test_update"
    ddb_cache.update_item({"id": "test_update"}, {"id": "test_update", "data": data})
    item = ddb_cache.get_item({"id": "test_update"})
    assert item
    assert item["id"] == "test_update"
    assert item["data"] == data


def test_delete_update(ddb_cache: DDBCache, init_update):
    data = randint(101, 200)  # noqa: S311
    item = ddb_cache.get_item({"id": "test_update"})
    assert item
    assert item["id"] == "test_update"
    assert item["data"] == "test_update"
    ddb_cache.del_update_item({"id": "test_update"}, {"id": "test_update", "data": data})
    item = ddb_cache.get_item({"id": "test_update"})
    assert item
    assert item["id"] == "test_update"
    assert item["data"] == data


def test_scan(ddb_cache: DDBCache, init_get, init_delete, init_update):
    items = ddb_cache.scan_items()
    assert len(items) >= 1


def test_query(ddb_cache: DDBCache, init_get, init_update, init_delete):
    items = ddb_cache.query_items({"KeyConditionExpression": Key("id").eq("test_get")})
    assert len(items) == 1
    items = ddb_cache.query_items({"KeyConditionExpression": Key("id").eq("test_update")})
    assert len(items) == 1
    items = ddb_cache.query_items({"KeyConditionExpression": Key("id").eq("test_delete")})
    assert len(items) == 1


def test_cache_floats(ddb_cache: DDBCache):
    data = float(2 / 3)
    input = {"id": "test_cache_floats", "data": data}
    print(input)
    ddb_cache.put_cache(input, with_ttl=False)
    item = ddb_cache.fetch_cache({"id": "test_cache_floats"})
    assert item
    assert item["id"] == "test_cache_floats"
    assert str(item["data"]) == str(data)
    assert type(item["data"]) is float
    item = ddb_cache.get_item({"id": "test_cache_floats"})
    assert item
    assert item["id"] == "test_cache_floats"
    assert str(item["data"]) == str(data)


def test_cache_floats_in_list(ddb_cache: DDBCache):
    data = float(2 / 3)
    ddb_cache.put_cache({"id": "test_cache_floats_in_list", "data": [data]}, with_ttl=False)
    item = ddb_cache.fetch_cache({"id": "test_cache_floats_in_list"})
    assert item
    assert item["id"] == "test_cache_floats_in_list"
    assert str(item["data"][0]) == str(data)
    assert type(item["data"][0]) is float
    item = ddb_cache.get_item({"id": "test_cache_floats_in_list"})
    assert item
    assert item["id"] == "test_cache_floats_in_list"
    assert str(item["data"][0]) == str(data)


def test_cache_floats_in_dict(ddb_cache: DDBCache):
    data = float(2 / 3)
    ddb_cache.put_cache({"id": "test_put_cache", "data": {"node": data}}, with_ttl=False)
    item = ddb_cache.fetch_cache({"id": "test_put_cache"})
    assert item
    assert item["id"] == "test_put_cache"
    assert str(item["data"]["node"]) == str(data)
    assert type(item["data"]["node"]) is float
    item = ddb_cache.get_item({"id": "test_put_cache"})
    assert item
    assert item["id"] == "test_put_cache"
    assert str(item["data"]["node"]) == str(data)


def test_cache_floats_in_tuples(ddb_cache: DDBCache):
    data = float(2 / 3)
    ddb_cache.put_cache({"id": "test_cache_floats_in_tuples", "data": ("val", data)}, with_ttl=False)
    item = ddb_cache.get_item({"id": "test_cache_floats_in_tuples"})
    assert item
    assert item["id"] == "test_cache_floats_in_tuples"
    assert str(item["data"][1]) == str(data)
    item = ddb_cache.fetch_cache({"id": "test_cache_floats_in_tuples"})
    assert item
    assert item["id"] == "test_cache_floats_in_tuples"
    assert str(item["data"][1]) == str(data)
    assert type(item["data"][1]) is float



# StdLib
from datetime import datetime
from decimal import Decimal
import json
from time import time
from typing import Any, Union

import boto3
from botocore.exceptions import ClientError
from helpers import Logger, NormalSleeper, RetryHandler

LOGGER = Logger()
SLEEPER = NormalSleeper()


def _client_error(error_obj):
    err_msg = str(error_obj)
    if "ProvisionedThroughputExceededException" not in err_msg:
        raise error_obj
    else:
        LOGGER.print_error(err_msg)
        SLEEPER.sleep(10)


class DDBCache:
    TTL_DAYS = 30

    def __init__(self, table_name, access_key, secret_key, region):
        self.table_name = table_name
        self._setup_session(access_key, secret_key, region)

    def _setup_session(self, access_key, secret_key, region):
        _session_config = {
            "region_name": region,
            "aws_access_key_id": access_key,
            "aws_secret_access_key": secret_key,
        }
        _service_config = {
            "service_name": "dynamodb",
        }
        self._SESSION = boto3.Session(**_session_config)
        self._DDB = self._SESSION.resource(**_service_config)
        self._DDB_CLIENT = self._SESSION.client(**_service_config)

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def create_table(
        self,
        key_schema: list[dict[str, Any]] = [],
        attr_schema: list[dict[str, Any]] = [],
        **create_kwargs,
    ):
        try:
            tbl = self._DDB.Table(self.table_name)
            tbl.reload()
            self.TABLE = tbl
        except:  # noqa
            print("Table does not exist. Creating...")
            tbl = self._DDB.create_table(
                TableName=self.table_name,
                KeySchema=key_schema,
                AttributeDefinitions=attr_schema,
                **create_kwargs,
            )
            self.TABLE = tbl

    def create_backup(self):
        response = self._DDB_CLIENT.create_backup(
            TableName=self.table_name,
            BackupName=datetime.now().isoformat().replace(":", "_"),
        )
        print(f"Backup: {response}")

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def put_item(self, data: dict[str, Any]):
        response = self.TABLE.put_item(Item=data)
        return response

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def get_item(self, key: dict[str, Any]) -> dict[str, Any]:
        response = self.TABLE.get_item(Key=key)
        return response.get("Item", None)

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def delete_item(self, key: dict[str, Any]):
        response = self.TABLE.delete_item(Key=key)
        return response

    def del_update_item(self, key: dict[str, Any], data: dict[str, Any]):
        self.delete_item(key)
        return self.put_item(data)

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def update_item(self, key_dict: dict[str, Any], data_dict: dict[str, Any]):

        data_formatted = {key: {"Value": value, "Action": "PUT"} for (key, value) in data_dict.items()}
        for k in key_dict.keys():
            del data_formatted[k]

        response = self.TABLE.update_item(
            Key=key_dict,
            AttributeUpdates=data_formatted,
            ReturnValues="ALL_NEW",
        )
        return response

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def scan_items(self) -> list[dict[str, Any]]:
        scan_kwargs: dict[str, Any] = {
            "Select": "ALL_ATTRIBUTES",
        }
        done = False
        start_key = None

        items: list = []

        while not done:
            if start_key:
                scan_kwargs["ExclusiveStartKey"] = start_key
            response = self.TABLE.scan(**scan_kwargs)
            for item in response.get("Items", []):
                items.append(item)
            start_key = response.get("LastEvaluatedKey", None)
            done = start_key is None

        return items

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def query_items(self, query_conditions) -> list[dict[str, Any]]:
        response = self.TABLE.query(**query_conditions)
        return response["Items"]

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def fetch_cache(self, key: dict[str, Any], with_ttl: bool = True) -> dict[str, Any]:
        cache = self.get_item(key)
        if cache and isinstance(cache, dict):
            if with_ttl:
                cache["ttl"] = self._get_item_ttl()
                self.update_item(key, cache)
            if cache.get("ttl", None):
                del cache["ttl"]
            return _decimal_to_float(cache)
        else:
            return {}

    @RetryHandler(
        (ClientError),
        max_retries=10,
        wait_time=0,
        err_callbacks={"ClientError": (_client_error, {})},
    ).wrap
    def put_cache(self, cache: dict[str, Any], with_ttl=True):
        cache_for_db = _float_to_decimal(cache)
        if with_ttl:
            cache_for_db["ttl"] = self._get_item_ttl()
        self.put_item(data=cache_for_db)
        if with_ttl and cache_for_db.get("ttl", None):
            del cache_for_db["ttl"]

    def _get_item_ttl(self):
        return int(time()) + int(60 * 60 * 24 * self.TTL_DAYS)


def _float_to_decimal(obj: Any):
    if not obj or type(obj) not in [dict, list]:
        return obj
    else:
        return json.loads(json.dumps(obj), parse_float=Decimal)


def _decimal_to_float(obj: Any):
    if not obj:
        return obj
    if type(obj) is Decimal:
        return float(obj)
    if type(obj) is dict:
        return {k: _decimal_to_float(v) for k, v in obj.items()}
    if type(obj) is list:
        return [_decimal_to_float(o) for o in obj]
    return obj
